/*
 * MPDclient - Music Player Daemon (MPD) remote for Mac OS
 * Copyright (C) 2019  Old-Geek
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <include/mpd/status.h>
#import "QueueView.h"
#import "../../Model/Song.h"
#import "../../Model/Stream.h"
#import "../../MPDService/Status/MPDStatus.h"
#import "../../MPDService/MPDCommand.h"
#import "../../MPDService/Helper/MPDQueue.h"
#import "../../MPDService/Helper/MPDStream.h"


@implementation QueueView

MPDStatus *mpdStatusQTV;

NSArray *queueArray;
NSImage *logoPlayQTV;

int oldPositionQTV;
long rowQueueSelected;

- (void)viewDidLoad {
    [super viewDidLoad];
    
    oldPositionQTV = [[MPDCache sMPDCache] getPosition];
    
    // purge sinon crash si nouvelle liste inférieur en nombre à l'ancienne
    queueArray = nil;
    [_queueTableView reloadData];
    
    queueArray = [[MPDQueue sMPDQueue] getQueue];
    [MPDCache sMPDCache].delegateQueue = self;
    [_queueTableView setTarget:self];
    [_queueTableView setDoubleAction:@selector(doubleClick:)];
    
    mpdStatusQTV = [[MPDCache sMPDCache] getStatus];
    [self onStateChange:mpdStatusQTV.state position:mpdStatusQTV.titlePosition];
}

- (void)awakeFromNib {
    [self performSelectorOnMainThread:@selector(updateTable:) withObject:_queueTableView waitUntilDone:NO];
}


- (void)updateTable:(NSTableView *)table {
    [table reloadData];
}


- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView {
    _queueTableView = tableView;
    return [queueArray count];
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row {
    Song *song = queueArray[row];

    mpdStatusQTV = [[MPDCache sMPDCache] getStatus];
    
    NSString *sTitle = mpdStatusQTV.title;
    NSString *sArtist = mpdStatusQTV.albumArtist;
    NSString *artist = song.albumArtist;
    
    if ([artist isEqualToString:@""]) {
        sArtist = mpdStatusQTV.artist;
        artist = song.artist;
    }
    
    return [self setCells:sArtist statusTitle:sTitle song:song tableColumn:tableColumn tableView:tableView artist:artist];
}


- (NSString *)setCells:(NSString *)sArtist statusTitle:(NSString *)sTitle song:(Song *)song tableColumn:(NSTableColumn * _Nullable)tableColumn tableView:(NSTableView * _Nonnull)tableView artist:(NSString *)artist {
    
    NSString *identifier = tableColumn.identifier;
    NSTableCellView *cell = [tableView makeViewWithIdentifier:identifier owner:self];
    
    if ([tableColumn.identifier isEqualToString:@"trackNumber"])
        return [self setTrack:cell statusArtist:sArtist statusTitle:sTitle song:song artist:artist];
    else if ([identifier isEqualToString:@"artist"])
        return cell.textField.stringValue = [NSString stringWithFormat:@"%@",artist];
    else if ([identifier isEqualToString:@"album"])
        return cell.textField.stringValue = [NSString stringWithFormat:@"%@",song.album];
    else if ([identifier isEqualToString:@"title"])
        return [self setTitle:cell song:song];
    else if ([identifier isEqualToString:@"year"])
        return cell.textField.stringValue = [NSString stringWithFormat:@"%d",song.date];
    else if ([identifier isEqualToString:@"time"])
        return [self setTime:cell statusArtist:sArtist statusTitle:sTitle song:song artist:artist];
    else
        return nil;
}


- (NSString *)setTrack:(NSTableCellView *)cell statusArtist:(NSString *)sArtist statusTitle:(NSString *)sTitle song:(Song *)song artist:(NSString *)artist {
    
    if ([sTitle isEqualToString:song.title] && [sArtist isEqualToString:artist] && (mpdStatusQTV.state != MPD_STATE_STOP)) {
        
        cell.imageView.image = logoPlayQTV;
        cell.textField.stringValue =@"";
        
    } else {
        
        cell.imageView.image = nil;;
        
        if (song.diskNb != 0)
            cell.textField.stringValue = [NSString stringWithFormat:@"%d ∙ %d",song.diskNb, song.track];
        else
            cell.textField.stringValue = [NSString stringWithFormat:@"%d", song.track];
    }
    
    return cell.textField.stringValue;
}

- (NSString *)setTitle:(NSTableCellView *)cell song:(Song *)song {
    
    NSString *path = song.fullPath;
    
    if (path.length > 4 && [[path substringToIndex:4] isEqualToString:@"http"]) {
        
        NSArray *streamArray =  [[MPDStream sMPDStream] getStreamList];
        
        for (Stream *stream in streamArray) {
            NSString *url = stream.url;
            
            if ([url isEqualToString:path])
                return cell.textField.stringValue = [NSString stringWithFormat:@"%@",stream.name];
        }
    }
    
    return cell.textField.stringValue = [NSString stringWithFormat:@"%@",song.title];
}


- (NSString *)setTime:(NSTableCellView *)cell statusArtist:(NSString *)statusArtist statusTitle:(NSString *)statusTitle song:(Song *)song artist:(NSString *)artist {
    
    int minutes = song.duration / 60;
    int seconds = (int)song.duration % 60;
    NSString *time = [NSString stringWithFormat:@"%2d:%02d",minutes, seconds];
    
    if ([statusTitle isEqualToString:song.title] && [statusArtist isEqualToString:artist] && (mpdStatusQTV.state != MPD_STATE_STOP))
        cell.textField.textColor = [NSColor colorWithRed:255.0/255.0 green:171.0/255.0 blue:1.0/255.0 alpha:1.0];
    else
        cell.textField.textColor = [NSColor secondaryLabelColor];
    
    return cell.textField.stringValue = [NSString stringWithFormat:@"%@",time];
}


- (void)doubleClick:(id)object {
    
    unsigned int rowNumber = (unsigned int) [_queueTableView clickedRow];
    
    [[MPDCommand sMPDCommand] playQueuePosition:rowNumber];
    
}


- (void)onStateChange:(int)state position:(int)position {
        
    if (state == MPD_STATE_PLAY)
        logoPlayQTV = [NSImage imageNamed:@"state_play"];
    else if (state == MPD_STATE_PAUSE)
        logoPlayQTV = [NSImage imageNamed:@"state_pause"];
    else
        logoPlayQTV = nil;
    
    [_queueTableView reloadDataForRowIndexes:[NSIndexSet indexSetWithIndex:position] columnIndexes:[NSIndexSet indexSetWithIndex:0]];
}


- (void)onTrackChange:(int)position {
    
    // colonne #track
    [_queueTableView reloadDataForRowIndexes:[NSIndexSet indexSetWithIndex:oldPositionQTV]
                     columnIndexes:[NSIndexSet indexSetWithIndex:0]];
    // colonne durée
    [_queueTableView reloadDataForRowIndexes:[NSIndexSet indexSetWithIndex:oldPositionQTV]
                               columnIndexes:[NSIndexSet indexSetWithIndex:4]];
    
    [_queueTableView reloadDataForRowIndexes:[NSIndexSet indexSetWithIndex:position]
                     columnIndexes:[NSIndexSet indexSetWithIndex:0]];
    
    [_queueTableView reloadDataForRowIndexes:[NSIndexSet indexSetWithIndex:position]
                               columnIndexes:[NSIndexSet indexSetWithIndex:4]];
    
    oldPositionQTV = position;
}


- (void)menuWillOpen:(NSMenu *)menu {
    rowQueueSelected = (long)_queueTableView.clickedRow;
}


- (IBAction)supprimer:(id)sender {
    
    NSIndexSet *row = [_queueTableView selectedRowIndexes];
    
    NSMutableArray *selectedItems = [NSMutableArray array];
    [row enumerateIndexesUsingBlock:^(NSUInteger idx, BOOL *stop) {
        [selectedItems addObject:[NSNumber numberWithInteger:idx]];
    }];
    
    if ([selectedItems count] > 1) {
        
        // évite à mpd de bug et de devoir relancer le service
        if ([queueArray count] == [selectedItems count])
            [[MPDCommand sMPDCommand] stop];
        
        [[MPDQueue sMPDQueue] removeInQueue:selectedItems];
        
    } else {
        
        if ([queueArray count] == 1)
            [[MPDCommand sMPDCommand] stop];
        
        NSMutableArray *selectedOne = [NSMutableArray array];
        [selectedOne addObject:[NSString stringWithFormat:@"%d", (int)rowQueueSelected]];
        
        [[MPDQueue sMPDQueue] removeInQueue:selectedOne];
    }
    
    queueArray = [[MPDQueue sMPDQueue] getQueue];
    [_queueTableView reloadData];
    
}


@end
